// EngineSchemaProvider (engine schema reg => schema)
import { NotificationHandler } from "vscode-languageserver/node";
import gql from "graphql-tag";
import { GraphQLSchema, IntrospectionQuery, buildClientSchema } from "graphql";
import { ApolloEngineClient, ClientIdentity } from "../../engine";
import { ClientConfig, keyEnvVar } from "../../config";
import {
  GraphQLSchemaProvider,
  SchemaChangeUnsubscribeHandler,
  SchemaResolveConfig,
} from "./base";

import {
  GetSchemaByTagQuery,
  GetSchemaByTagQueryVariables,
} from "../../graphqlTypes";
import { Debug } from "../../utilities";
import { TypedDocumentNode } from "@apollo/client/core";

export class EngineSchemaProvider implements GraphQLSchemaProvider {
  private schema?: GraphQLSchema;
  private client?: ApolloEngineClient;

  constructor(
    private config: ClientConfig,
    private clientIdentity: ClientIdentity,
  ) {}

  async resolveSchema(override: SchemaResolveConfig) {
    if (this.schema && (!override || !override.force)) return this.schema;
    const { engine, client } = this.config;

    if (!this.config.graph) {
      throw new Error(
        `No graph ID found for client. Please specify a graph ID via the config or the --graph flag`,
      );
    }

    // create engine client
    if (!this.client) {
      if (!engine.apiKey) {
        throw new Error(
          `No API key found. Please set ${keyEnvVar} or use --key`,
        );
      }
      this.client = new ApolloEngineClient(
        engine.apiKey,
        engine.endpoint,
        this.clientIdentity,
      );
    }

    const { data, errors } = await this.client.query({
      query: SCHEMA_QUERY,
      variables: {
        id: this.config.graph,
        tag: override && override.tag ? override.tag : this.config.variant,
      },
      fetchPolicy: "no-cache",
    });
    if (errors) {
      // XXX better error handling of GraphQL errors
      throw new Error(errors.map(({ message }) => message).join("\n"));
    }

    if (!(data && data.service && data.service.__typename === "Service")) {
      throw new Error(
        `Unable to get schema from the Apollo registry for graph ${this.config.graph}`,
      );
    }

    this.schema = buildClientSchema(
      data.service.schema as unknown as IntrospectionQuery,
    );
    return this.schema;
  }

  onSchemaChange(
    _handler: NotificationHandler<GraphQLSchema>,
  ): SchemaChangeUnsubscribeHandler {
    throw new Error("Polling of Apollo not implemented yet");
    return () => {};
  }

  async resolveFederatedServiceSDL() {
    Debug.error(
      "Cannot resolve a federated service's SDL from Apollo. Use an endpoint or a file instead",
    );
    return;
  }
}

export const SCHEMA_QUERY: TypedDocumentNode<
  GetSchemaByTagQuery,
  GetSchemaByTagQueryVariables
> = gql`
  query GetSchemaByTag($tag: String!, $id: ID!) {
    service(id: $id) {
      ... on Service {
        __typename
        schema(tag: $tag) {
          hash
          __schema: introspection {
            queryType {
              name
            }
            mutationType {
              name
            }
            subscriptionType {
              name
            }
            types(filter: { includeBuiltInTypes: true }) {
              ...IntrospectionFullType
            }
            directives {
              name
              description
              locations
              args {
                ...IntrospectionInputValue
              }
            }
          }
        }
      }
    }
  }

  fragment IntrospectionFullType on IntrospectionType {
    kind
    name
    description
    fields {
      name
      description
      args {
        ...IntrospectionInputValue
      }
      type {
        ...IntrospectionTypeRef
      }
      isDeprecated
      deprecationReason
    }
    inputFields {
      ...IntrospectionInputValue
    }
    interfaces {
      ...IntrospectionTypeRef
    }
    enumValues(includeDeprecated: true) {
      name
      description
      isDeprecated
      deprecationReason
    }
    possibleTypes {
      ...IntrospectionTypeRef
    }
  }

  fragment IntrospectionInputValue on IntrospectionInputValue {
    name
    description
    type {
      ...IntrospectionTypeRef
    }
    defaultValue
  }

  fragment IntrospectionTypeRef on IntrospectionType {
    kind
    name
    ofType {
      kind
      name
      ofType {
        kind
        name
        ofType {
          kind
          name
          ofType {
            kind
            name
            ofType {
              kind
              name
              ofType {
                kind
                name
                ofType {
                  kind
                  name
                }
              }
            }
          }
        }
      }
    }
  }
`;
