// @ts-check
async function loadDefaultMocks(/** @type {string} */ baseUri) {
  await sendMock(baseUri, FrontendUrlRoot);
  await sendMock(baseUri, SchemaTagsAndFieldStats);
  await sendMock(baseUri, GetSchemaByTag);
}

function sendMock(
  /** @type {string} */ baseUri,
  /** @type { { operationName: string, variables: Record<string, string>, response: unknown }} */ {
    operationName,
    variables,
    response,
  },
) {
  return require("undici").fetch(`${baseUri}/apollo`, {
    method: "PUT",
    body: JSON.stringify({
      operationName,
      variables,
      response,
    }),
    dispatcher: new (require("undici").Agent)({
      connect: {
        rejectUnauthorized: false,
      },
    }),
  });
}

const FrontendUrlRoot = {
  operationName: "FrontendUrlRoot",
  variables: {},
  response: { data: { frontendUrlRoot: "https://studio.apollographql.com" } },
};

const SchemaTagsAndFieldStats = {
  operationName: "SchemaTagsAndFieldStats",
  variables: { id: "spotify-demo-graph-519427f5" },
  response: {
    data: {
      service: {
        schemaTags: [
          { tag: "main", __typename: "SchemaTag" },
          { tag: "public", __typename: "SchemaTag" },
        ],
        stats: {
          fieldLatencies: [
            {
              __typename: "ServiceFieldLatenciesRecord",
              groupBy: {
                __typename: "ServiceFieldLatenciesDimensions",
                parentType: "CurrentUserProfile",
                fieldName: "id",
              },
              metrics: {
                __typename: "ServiceFieldLatenciesMetrics",
                fieldHistogram: {
                  __typename: "DurationHistogram",
                  durationMs: 1.2,
                },
              },
            },
            {
              __typename: "ServiceFieldLatenciesRecord",
              groupBy: {
                __typename: "ServiceFieldLatenciesDimensions",
                parentType: "CurrentUserProfile",
                fieldName: "displayName",
              },
              metrics: {
                __typename: "ServiceFieldLatenciesMetrics",
                fieldHistogram: {
                  __typename: "DurationHistogram",
                  durationMs: 12.8,
                },
              },
            },
            {
              __typename: "ServiceFieldLatenciesRecord",
              groupBy: {
                __typename: "ServiceFieldLatenciesDimensions",
                parentType: "CurrentUser",
                fieldName: "profile",
              },
              metrics: {
                __typename: "ServiceFieldLatenciesMetrics",
                fieldHistogram: {
                  __typename: "DurationHistogram",
                  durationMs: 14.5,
                },
              },
            },
            {
              groupBy: {
                parentType: "Query",
                fieldName: "me",
                __typename: "ServiceFieldLatenciesDimensions",
              },
              metrics: {
                fieldHistogram: {
                  durationMs: 83.89787623554304,
                  __typename: "DurationHistogram",
                },
                __typename: "ServiceFieldLatenciesMetrics",
              },
              __typename: "ServiceFieldLatenciesRecord",
            },
          ],
          __typename: "ServiceStatsWindow",
        },
        __typename: "Service",
      },
    },
  },
};
const SchemaTagsAndFieldStats_WRONG_TOKEN = {
  operationName: "SchemaTagsAndFieldStats",
  variables: { id: "spotify-demo-graph-519427f5" },
  response: {
    data: { service: null },
    errors: [
      {
        message: "HTTP fetch failed from 'kotlin': 406: Not Acceptable",
        extensions: {
          code: "SUBREQUEST_HTTP_ERROR",
          service: "kotlin",
          reason: "406: Not Acceptable",
          http: { status: 406 },
        },
      },
      { message: "Invalid credentials provided" },
    ],
  },
};
const SchemaTagsAndFieldStats_NOT_ALLOWED_BY_USER_ROLE = {
  operationName: "SchemaTagsAndFieldStats",
  variables: { id: "spotify-demo-graph-519427f5" },
  response: {
    data: { service: null },
    errors: [
      {
        message:
          "Exception while fetching data (/service/stats) : NOT_ALLOWED_BY_USER_ROLE",
        locations: [{ line: 1, column: 103 }],
        path: ["service", "stats"],
        extensions: { classification: "DataFetchingException" },
      },
    ],
  },
};
const GetSchemaByTag = {
  operationName: "GetSchemaByTag",
  variables: { id: "spotify-demo-graph-519427f5", tag: "main" },
  response: {
    data: {
      service: {
        __typename: "Service",
        schema: {
          hash: "ae186ecdb3cab1b7acc6ea37b704e7e83dc77631ac9efa86015a66973b5de3bb",
          __schema: {
            queryType: { name: "Query", __typename: "IntrospectionType" },
            mutationType: { name: "Mutation", __typename: "IntrospectionType" },
            subscriptionType: {
              name: "Subscription",
              __typename: "IntrospectionType",
            },
            types: [
              {
                kind: "ENUM",
                name: "Action",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "INTERRUPTING_PLAYBACK",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "PAUSING",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "RESUMING",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "SEEKING",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "SKIPPING_NEXT",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "SKIPPING_PREV",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "TOGGLING_REPEAT_CONTEXT",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "TOGGLING_SHUFFLE",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "TOGGLING_REPEAT_TRACK",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "TRANSFERRING_PLAYBACK",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Actions",
                description: null,
                fields: [
                  {
                    name: "disallows",
                    description: null,
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "ENUM",
                            name: "Action",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "AddItemsToPlaylistInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "playlistId",
                    description:
                      "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nof the playlist.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "uris",
                    description:
                      "A comma-separated list of [Spotify URIs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nto add, can be track or episode URIs. A maximum of 100 items can be added in\none request.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "String",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "position",
                    description:
                      "The position to insert the items, a zero-based index. For example, to insert\nthe items in the first position: **position=0**; to insert the items in the\nthird position: **position=2**. If omitted, the items will be appended to the\nplaylist. Items are added in the order they are listed in the query string or\nrequest body.",
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "SCALAR",
                name: "ID",
                description:
                  'The `ID` scalar type represents a unique identifier, often used to refetch an object or as key for a cache. The ID type appears in a JSON response as a String; however, it is not intended to be human-readable. When expected as an input type, any string (such as `"4"`) or integer (such as `4`) input value will be accepted as an ID.',
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "SCALAR",
                name: "String",
                description:
                  "The `String` scalar type represents textual data, represented as UTF-8 character sequences. The String type is most often used by GraphQL to represent free-form human-readable text.",
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "SCALAR",
                name: "Int",
                description:
                  "The `Int` scalar type represents non-fractional signed whole numeric values. Int can represent values between -(2^31) and 2^31 - 1.",
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "AddItemsToPlaylistPayload",
                description: null,
                fields: [
                  {
                    name: "playlist",
                    description:
                      "The playlist that contains the newly added items",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "Playlist",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "AddItemToPlaybackQueueInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "uri",
                    description:
                      "The uri of the item to add to the queue. Must be a track or an episode uri.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "deviceId",
                    description:
                      "The id of the device this command is targeting. If not supplied, the user's\ncurrently active device is the target.",
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "AddItemToPlaybackQueuePayload",
                description: null,
                fields: [
                  {
                    name: "playbackQueue",
                    description: null,
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackQueue",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Album",
                description: "Spotify catalog information for an album.",
                fields: [
                  {
                    name: "albumType",
                    description: "The type of the album.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "ENUM",
                        name: "AlbumType",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "artists",
                    description: "The artists of the album.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "Artist",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "copyrights",
                    description: "The copyrights for the album.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "Copyright",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalUrls",
                    description: "Known external URLs for this album.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExternalUrl",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "genres",
                    description: "Genres for the album.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "String",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint providing full details of the album.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "id",
                    description:
                      "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the album.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "images",
                    description:
                      "The cover art for the album in various sizes, widest first.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "Image",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "label",
                    description: "The label the album was released under.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "name",
                    description:
                      "The name of the album. In case of an album takedown, the value may be an empty\nstring.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "releaseDate",
                    description: "The date the album was first released.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ReleaseDate",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "totalTracks",
                    description: "The number of tracks in the album.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "tracks",
                    description: "The tracks of the album.",
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first playlist to return. Default: 0 (the first object).\n\nUse with `limit` to get the next set of tracks.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "AlbumTrackConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the album.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "AlbumGroup",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "ALBUM",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "SINGLE",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "APPEARS_ON",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "COMPILATION",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "AlbumTrackConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The set of tracks.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "AlbumTrackEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for the set of tracks.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "AlbumTrackEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The track on the album",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Track",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "AlbumType",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "ALBUM",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "SINGLE",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "COMPILATION",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Artist",
                description: "Spotify catalog information for an artist.",
                fields: [
                  {
                    name: "albums",
                    description:
                      "Spotify catalog information about an artist's albums.",
                    args: [
                      {
                        name: "includeGroups",
                        description:
                          "Used to filter the response. If not supplied, all album types will be\nreturned.",
                        type: {
                          kind: "LIST",
                          name: null,
                          ofType: {
                            kind: "NON_NULL",
                            name: null,
                            ofType: {
                              kind: "ENUM",
                              name: "AlbumGroup",
                              ofType: null,
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first item to return. Default: 0 (the first item). Use\nwith `limit` to get the next set of items.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "ArtistAlbumsConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalUrls",
                    description: "Known external URLs for this artist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExternalUrl",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "followers",
                    description:
                      "Information about the followers of the artist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Followers",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "genres",
                    description:
                      "A list of the genres the artist is associated with. If not yet classified, the\narray is empty.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "String",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint providing full details of the artist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "id",
                    description:
                      "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the artist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "images",
                    description:
                      "Images of the artist in various sizes, widest first.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "Image",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "name",
                    description: "The name of the artist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "popularity",
                    description:
                      "The popularity of the artist. The value will be between 0 and 100, with 100\nbeing the most popular. The artist's popularity is calculated from the\npopularity of all the artist's tracks.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "relatedArtists",
                    description:
                      "Spotify catalog information about artists similar to a given artist.\nSimilarity is based on analysis of the Spotify community's\n[listening history](http://news.spotify.com/se/2010/02/03/related-artists/).",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "Artist",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "topTracks",
                    description:
                      "Spotify catalog information about an artist's top tracks.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "Track",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the artist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ArtistAlbumEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "Spotify catalog information for the album.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Album",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ArtistAlbumsConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "A list of albums that belong to the artist.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "ArtistAlbumEdge",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      '"Pagination information for the set of albums"',
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Contains",
                description: null,
                fields: [
                  {
                    name: "albums",
                    description:
                      "List of booleans in order of albums requested. `true` means the album is in\nthe Spotify user's library. This field is `null` if omitted in the request.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "Boolean",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "episodes",
                    description:
                      "List of booleans in order of episodes requested. `true` means the episode is in\nthe Spotify user's library. This field is `null` if omitted in the request.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "Boolean",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "shows",
                    description:
                      "List of booleans in order of shows requested. `true` means the show is in\nthe Spotify user's library. This field is `null` if omitted in the request.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "Boolean",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "tracks",
                    description:
                      "List of booleans in order of tracks requested. `true` means the track is in\nthe Spotify user's library. This field is `null` if omitted in the request.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "Boolean",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "SCALAR",
                name: "Boolean",
                description:
                  "The `Boolean` scalar type represents `true` or `false`.",
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Copyright",
                description: null,
                fields: [
                  {
                    name: "text",
                    description: "The copyright text for this content.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "type",
                    description:
                      "The type of copyright: `C` = the copyright, `P` = the sound recording\n(performance) copyright.",
                    args: [],
                    type: {
                      kind: "ENUM",
                      name: "CopyrightType",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "CopyrightType",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "C",
                    description: "The copyright",
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "P",
                    description: "The sound recording (performance) copyright.",
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "SCALAR",
                name: "CountryCode",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "CurrentlyPlaying",
                description: null,
                fields: [
                  {
                    name: "actions",
                    description:
                      "Allows to update the user interface based on which playback actions are\navailable within the current context.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Actions",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "context",
                    description: "A context object.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackContext",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "isPlaying",
                    description:
                      "If something is currently playing, return `true`.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "item",
                    description: "The currently playing track or episode",
                    args: [],
                    type: {
                      kind: "INTERFACE",
                      name: "PlaybackItem",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "progressMs",
                    description:
                      "Progress into the currently playing track or episode. Can be `null`",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "timestamp",
                    description:
                      "Unix Millisecond Timestamp when data was fetched.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Timestamp",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "CurrentUser",
                description: null,
                fields: [
                  {
                    name: "albums",
                    description:
                      "Get a list of the albums saved in the current Spotify user's 'Your Music'\nlibrary.",
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first item to return. Default: 0 (the first item). Use with\nlimit to get the next set of items.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SavedAlbumsConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "albumsContains",
                    description:
                      "Check if one or more albums is already saved in the current Spotify user's\n'Your Music' library.",
                    args: [
                      {
                        name: "ids",
                        description:
                          "A comma-separated list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the albums. Maximum: 20 IDs.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "SCALAR",
                                name: "ID",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "Boolean",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "episodes",
                    description: null,
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first item to return. Default: 0 (the first item). Use with `limit` to get the next set of items.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SavedEpisodesConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "episodesContains",
                    description:
                      "Check if one or more episodes is already saved in the current Spotify user's\n'Your Episodes' library.",
                    args: [
                      {
                        name: "ids",
                        description:
                          "A comma-separated list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the episodes. Maximum: 50 IDs.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "SCALAR",
                                name: "ID",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "Boolean",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "followedArtists",
                    description: "Get the current user's followed artists.",
                    args: [
                      {
                        name: "after",
                        description:
                          "The last artist ID retrieved from the previous request.",
                        type: {
                          kind: "SCALAR",
                          name: "String",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "FollowedArtistsConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "user",
                    description:
                      "Detailed profile information about the current user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "User",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: true,
                    deprecationReason:
                      "Use the profile field instead which provides richer current user information.",
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "player",
                    description:
                      "Information about the user's current playback state",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Player",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "playlists",
                    description:
                      "Playlists owned or followed by the current Spotify user.",
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first playlist to return. Default: 0 (the first object).\n\nUse with `limit` to get the next set of playlists.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "PlaylistConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "profile",
                    description:
                      "Get detailed profile information about the current user (including the current user's username).",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "CurrentUserProfile",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "shows",
                    description:
                      "Get a list of the albums saved in the current Spotify user's 'Your Music' library.",
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first item to return. Default: 0 (the first item). Use with `limit` to get the next set of items.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SavedShowsConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "showsContains",
                    description:
                      "Check if one or more shows is already saved in the current Spotify user's\nlibrary.",
                    args: [
                      {
                        name: "ids",
                        description:
                          "A comma-separated list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the shows. Maximum: 50 IDs.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "SCALAR",
                                name: "ID",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "Boolean",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "topArtists",
                    description:
                      "Get the current user's top artists based on calculated affinity.",
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first item to return. Default: 0 (the first item). Use with limit to get the next set of items.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "timeRange",
                        description:
                          "Over what time frame the affinities are computed. Valid values: `LONG_TERM` (calculated from several years of data and including all new data as it becomes available), `MEDIUM_TERM` (approximately last 6 months), `SHORT_TERM` (approximately last 4 weeks). Default: `MEDIUM_TERM`",
                        type: {
                          kind: "ENUM",
                          name: "TimeRange",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "TopArtistsConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "topTracks",
                    description:
                      "Get the current user's top tracks based on calculated affinity.",
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first item to return. Default: 0 (the first item). Use with limit to get the next set of items.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "timeRange",
                        description:
                          "Over what time frame the affinities are computed. Valid values: `LONG_TERM` (calculated from several years of data and including all new data as it becomes available), `MEDIUM_TERM` (approximately last 6 months), `SHORT_TERM` (approximately last 4 weeks). Default: `MEDIUM_TERM`",
                        type: {
                          kind: "ENUM",
                          name: "TimeRange",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "TopTracksConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "tracks",
                    description: null,
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first track to return. Default: 0 (the first object).\n\nUse with `limit` to get the next set of tracks.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SavedTracksConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "tracksContains",
                    description:
                      "Check if one or more tracks is already saved in the current Spotify user's\n'Your Music' library.",
                    args: [
                      {
                        name: "ids",
                        description:
                          "A comma-separated list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the tracks. Maximum: 50 IDs.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "SCALAR",
                                name: "ID",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "Boolean",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "CurrentUserProfile",
                description: null,
                fields: [
                  {
                    name: "country",
                    description:
                      "The country of the user, as set in the user's account profile. An ISO 3166-1\nalpha-2 country code.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "CountryCode",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "displayName",
                    description:
                      "The name displayed on the user's profile. `null` if not available.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "email",
                    description:
                      "The user's email address, as entered by the user when creating their account.\n_**Important!** This email address is unverified; there is no proof that it\nactually belongs to the user._",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "explicitContent",
                    description: "The user's explicit content settings.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExplicitContentSettings",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "followers",
                    description: "Information about the followers of the user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Followers",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint for this user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "id",
                    description:
                      "The [Spotify user ID](https://developer.spotify.com/documentation/web-api/concepts/spotify-uris-ids)\nfor the user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "images",
                    description: "The user's profile image.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Image",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "product",
                    description:
                      'The user\'s Spotify subscription level: "premium", "free", etc. (The\nsubscription level "open" can be considered the same as "free".)',
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/concepts/spotify-uris-ids)\nfor the user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [
                  {
                    kind: "INTERFACE",
                    name: "UserProfile",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                ],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Cursors",
                description: null,
                fields: [
                  {
                    name: "after",
                    description:
                      "The cursor to use as key to find the next page of items.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "before",
                    description:
                      "The ursor to use as key to find the previous page of items.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "SCALAR",
                name: "DateTime",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Developer",
                description: null,
                fields: [
                  {
                    name: "fieldConfigs",
                    description:
                      "A list of configured GraphQL fields. Only fields that have non-zero timeouts\nand error rates will be listed.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "FieldConfig",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Device",
                description: null,
                fields: [
                  {
                    name: "id",
                    description: "The device ID",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "isActive",
                    description:
                      "If this device is the currently active device.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "isPrivateSession",
                    description:
                      "If this device is currently in a private session.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "isRestricted",
                    description:
                      'Whether controlling this device is restricted. At present if this is "true",\nthen no Web API commands will be accepted by this device.',
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "name",
                    description:
                      'A human-readable name for the device. Some devices have a name that the user\ncan configure (e.g. "Loudest speaker") and some devices have a generic name\nassociated with the manufacturer or device model.',
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "type",
                    description:
                      'Device type, such as "computer", "smartphone" or "speaker".',
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "volumePercent",
                    description:
                      "The current volume in percent.\n\n>= 0    <= 100",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Episode",
                description: "Spotify catalog information for an episode.",
                fields: [
                  {
                    name: "id",
                    description:
                      "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for the episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "audioPreviewUrl",
                    description:
                      "A URL to a 30 second preview (MP3 format) of the episode. `null` if not available.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "description",
                    description: "A description of the episode",
                    args: [
                      {
                        name: "format",
                        description: null,
                        type: {
                          kind: "ENUM",
                          name: "TextFormat",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: "PLAIN",
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "durationMs",
                    description: "The episode length in milliseconds.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "explicit",
                    description:
                      "Whether or not the episode has explicit content (`true` = yes it does;\n`false` = no it does not OR unknown).",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalUrls",
                    description: "External URLs for this episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExternalUrl",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint providing full details of the episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "images",
                    description:
                      "The cover art for the episode in various sizes, widest first.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "Image",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "isExternallyHosted",
                    description:
                      "`true` if the episode is hosted outside of Spotify's CDN.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "isPlayable",
                    description:
                      "`true` if the episode is playable in the given market. Otherwise `false`.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "languages",
                    description:
                      "A list of the languages used in the episode, identified by their\n[ISO 639-1](https://en.wikipedia.org/wiki/ISO_639) code.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "String",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "name",
                    description: "The name of the episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "releaseDate",
                    description: "The date the episode was first released",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ReleaseDate",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "resumePoint",
                    description:
                      "The user's most recent position in the episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ResumePoint",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "show",
                    description: "The show containing the episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Show",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [
                  {
                    kind: "INTERFACE",
                    name: "PlaylistTrack",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                  {
                    kind: "INTERFACE",
                    name: "PlaybackItem",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                ],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "SCALAR",
                name: "ErrorRate",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ExplicitContentSettings",
                description: null,
                fields: [
                  {
                    name: "filterEnabled",
                    description:
                      "When `true`, indicates that explicit content should not be played.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "filterLocked",
                    description:
                      "When `true`, indicates that the explicit content setting is locked and can't\nbe changed by the user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ExternalUrl",
                description: null,
                fields: [
                  {
                    name: "spotify",
                    description:
                      "The [Spotify URL](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the object.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "FeaturedPlaylistConnection",
                description: null,
                fields: [
                  {
                    name: "message",
                    description: null,
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "edges",
                    description:
                      "A list of Spotify featured playlists (shown, for example, on a Spotify player's\n'Browse' tab).",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "FeaturedPlaylistEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for the set of playlists",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "FeaturedPlaylistEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: null,
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Playlist",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "FieldConfig",
                description: null,
                fields: [
                  {
                    name: "schemaField",
                    description: "The schema field that includes this config",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "SchemaField",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "timeout",
                    description:
                      "The synthetic timeout configured for the field.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "errorRate",
                    description:
                      "The synthetic error rate configured for the field.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ErrorRate",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "FieldConfigInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "timeout",
                    description:
                      "The synthetic timeout configured for a field. Set to `null` to reset the value\nback to its default. Omit this field to maintain its value. Defaults to `0`.",
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "errorRate",
                    description:
                      "The synthetic error rate configured for a field. This should be a value\nbetween `0` and `1` where `0` means no synthetic errors should be thrown and\n`1` means errors should be thrown 100% of the time. Set to `null` to reset the\nvalue back to its default. Omit this field to maintain its value. Defaults to\n`0`.",
                    type: {
                      kind: "SCALAR",
                      name: "ErrorRate",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "FieldInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "schemaField",
                    description:
                      "Configure a field by its type in the schema. This will apply the config to all\nfields of the given type regardless of where it is queried in the scheam.\n\nOne of `path` or `schema` is required. If both are provided, `schema` will\ntake precendence as it has broader impact.",
                    type: {
                      kind: "INPUT_OBJECT",
                      name: "SchemaFieldInput",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "FollowedArtistEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The followed artist",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Artist",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "FollowedArtistsConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of followed artists.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "FollowedArtistEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for the set of followed artists.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfoCursorBased",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Followers",
                description: null,
                fields: [
                  {
                    name: "total",
                    description: "The total number of followers.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Image",
                description: null,
                fields: [
                  {
                    name: "url",
                    description: "The source URL of the image.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "height",
                    description: "The image height in pixels.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "width",
                    description: "The image width in pixels.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Mutation",
                description: null,
                fields: [
                  {
                    name: "resumePlayback",
                    description:
                      "Start a new context or resume current playback on the user's active device.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "ResumePlaybackInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "ResumePlaybackPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pausePlayback",
                    description: "Pause playback on the user's account.",
                    args: [
                      {
                        name: "context",
                        description:
                          "Additional context when pausing playback. Omit to pause playback on the\ncurrent device.",
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "PausePlaybackContextInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "PausePlaybackResponse",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "seekToPosition",
                    description:
                      "Seeks to the given position in the user’s currently playing track.",
                    args: [
                      {
                        name: "positionMs",
                        description:
                          "The position in milliseconds to seek to. Must be a positive number. Passing\nin a position that is greater than the length of the track will cause the\nplayer to start playing the next song.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "Int",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "context",
                        description:
                          "Additional context to determine the device where the action should occur.",
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "SeekToPositionContextInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SeekToPositionResponse",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "setRepeatMode",
                    description: "Set the repeat mode for the user's playback.",
                    args: [
                      {
                        name: "state",
                        description:
                          "`track`, `context` or `off`.\n`track` will repeat the current track.\n`context` will repeat the current context.\n`off` will turn repeat off.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "ENUM",
                            name: "RepeatMode",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "context",
                        description:
                          "Additional context to determine the device where the action should occur.",
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "SetRepeatModeContextInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SetRepeatModeResponse",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "setVolume",
                    description:
                      "Set the volume for the user’s current playback device.",
                    args: [
                      {
                        name: "volumePercent",
                        description:
                          "The volume to set. Must be a value from 0 to 100 inclusive.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "Int",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "context",
                        description:
                          "Additional context to determine the device where the action should occur.",
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "SetVolumeContextInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SetVolumeResponse",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "shufflePlayback",
                    description:
                      "Toggle shuffle on or off for user’s playback.",
                    args: [
                      {
                        name: "state",
                        description:
                          "`true`: Shuffle user's playback.\n`false`: Do not shuffle user's playback.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "Boolean",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "context",
                        description:
                          "Additional context to determine the device where the action should occur.",
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "ShufflePlaybackContextInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "ShufflePlaybackResponse",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "skipToNext",
                    description: "Skips to next track in the user’s queue.",
                    args: [
                      {
                        name: "context",
                        description:
                          "Additional context to determine the device where the action should occur.",
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "SkipToNextContextInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SkipToNextResponse",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "skipToPrevious",
                    description: "Skips to previous track in the user’s queue.",
                    args: [
                      {
                        name: "context",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "SkipToPreviousContextInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SkipToPreviousResponse",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "transferPlayback",
                    description:
                      "Transfer playback to a new device and determine if it should start playing.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "TransferPlaybackInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "TransferPlaybackPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "addItemToPlaybackQueue",
                    description:
                      "Add an item to the end of the user's current playback queue.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "AddItemToPlaybackQueueInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "AddItemToPlaybackQueuePayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "addItemsToPlaylist",
                    description: "Add one or more items to a user's playlist.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "AddItemsToPlaylistInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "AddItemsToPlaylistPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "updateFieldConfig",
                    description:
                      "Update configuration for a field in the schema. Allows tweaks to the\nsynthetic timeouts and error rates associated with the field. By default, both\nthe timeout and error rate are set to 0.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "UpdateFieldConfigInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "UpdateFieldConfigPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "removeItemFromPlaylist",
                    description:
                      "Remove one or more items from a user's playlist.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "RemoveItemFromPlaylistInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "RemoveItemFromPlaylistPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "removeSavedAlbums",
                    description:
                      "Remove one or more albums from the current user's 'Your Music' library.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "RemoveSavedAlbumsInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "RemoveSavedAlbumsPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "removeSavedEpisodes",
                    description:
                      "Remove one or more episodes from the current user's library.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "RemoveSavedEpisodesInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "RemoveSavedEpisodesPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "removeSavedShows",
                    description:
                      "Delete one or more shows from current Spotify user's library.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "RemoveSavedShowsInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "RemoveSavedShowsPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "removeSavedTracks",
                    description:
                      "Remove one or more tracks from the current user's 'Your Music' library.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "RemoveSavedTracksInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "RemoveSavedTracksPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "resetFieldConfig",
                    description:
                      "Reset a field's config back to its default values.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "ResetFieldConfigInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "ResetFieldConfigPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "saveAlbums",
                    description:
                      "Save one or more albums to the current user's 'Your Music' library.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "SaveAlbumsInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SaveAlbumsPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "saveEpisodes",
                    description:
                      "Save one or more episodes to the current user's library.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "SaveEpisodesInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SaveEpisodesPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "saveShows",
                    description:
                      "Save one or more shows to current Spotify user's library.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "SaveShowsInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SaveShowsPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "saveTracks",
                    description:
                      "Save one or more tracks to the current user's 'Your Music' library.",
                    args: [
                      {
                        name: "input",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "SaveTracksInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SaveTracksPayload",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "NewReleaseEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The newly released album",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Album",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "NewReleasesConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of new releases",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "NewReleaseEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description: "Pagination information for the new releases",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "PageInfo",
                description: null,
                fields: [
                  {
                    name: "hasNextPage",
                    description: "Whether there is a next page of items.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "hasPreviousPage",
                    description: "Whether there is a previous page of items.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "limit",
                    description:
                      "The maximum number of items in the response (as set in the query or default)",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "offset",
                    description:
                      "The offset of the items returned (as set in the query or default)",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "total",
                    description:
                      "The total number of items returned for the page.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "PageInfoCursorBased",
                description: null,
                fields: [
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint returning the full result of the request.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "limit",
                    description:
                      "The maximum number of items in the response (as set in the query or default)",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "next",
                    description:
                      "URL to the next page of items. (`null` if none)",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "cursors",
                    description:
                      "The cursors used to find the next set of items.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "Cursors",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "total",
                    description:
                      "The total number of items available to return.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "PausePlaybackContextInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "deviceId",
                    description:
                      "The id of the device this command is targeting. If not supplied, the user's\ncurrently active device is the target.",
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "PausePlaybackResponse",
                description: null,
                fields: [
                  {
                    name: "playbackState",
                    description: "The updated playback state",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "PlaybackContext",
                description: null,
                fields: [
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint providing full details of the track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "type",
                    description:
                      'The object type, e.g. "artist", "playlist", "album", "show".',
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "ENUM",
                        name: "PlaybackContextType",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalUrls",
                    description: "External URLs for this context.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExternalUrl",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the context.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "PlaybackContextType",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "ALBUM",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "ARTIST",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "AUDIO_FEATURES",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "COLLECTION",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "COLLECTION_YOUR_EPISODES",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "EPISODE",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "GENRE",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "PLAYLIST",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "SHOW",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "TRACK",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "USER",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INTERFACE",
                name: "PlaybackItem",
                description: null,
                fields: [
                  {
                    name: "id",
                    description:
                      "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the playback item.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "durationMs",
                    description:
                      "The duration for the playback item in milliseconds.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalUrls",
                    description: "Known external URLs for this playback item.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExternalUrl",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint providing full details of the playlist item.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "name",
                    description: "The name of the playlist item.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: [
                  {
                    kind: "OBJECT",
                    name: "Episode",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                  {
                    kind: "OBJECT",
                    name: "Track",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                ],
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "PlaybackQueue",
                description: null,
                fields: [
                  {
                    name: "currentlyPlaying",
                    description: null,
                    args: [],
                    type: {
                      kind: "INTERFACE",
                      name: "PlaybackItem",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "queue",
                    description: null,
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INTERFACE",
                            name: "PlaybackItem",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "PlaybackState",
                description: null,
                fields: [
                  {
                    name: "actions",
                    description:
                      "Allows to update the user interface based on which playback actions are\navailable within the current context.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Actions",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "context",
                    description: "A context object.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackContext",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "device",
                    description: "The device that is currently active.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Device",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "isPlaying",
                    description:
                      "If something is currently playing, return `true`.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "item",
                    description: "The currently playing track or episode",
                    args: [],
                    type: {
                      kind: "INTERFACE",
                      name: "PlaybackItem",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "progressMs",
                    description:
                      "Progress into the currently playing track or episode. Can be `null`",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "repeatState",
                    description: "off, track, context",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "ENUM",
                        name: "RepeatMode",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "shuffleState",
                    description: "If shuffle is on or off.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "timestamp",
                    description:
                      "Unix Millisecond Timestamp when data was fetched.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Timestamp",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Player",
                description: null,
                fields: [
                  {
                    name: "currentlyPlaying",
                    description:
                      "Information about the object currently being played on the user's Spotify account.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "CurrentlyPlaying",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "devices",
                    description:
                      "Information about a user's available devices.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Device",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "playbackQueue",
                    description:
                      "Get the list of objects that make up the user's queue.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackQueue",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "playbackState",
                    description:
                      "Information about the user's current playback state, including track or\nepisode, progress, and active device.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "recentlyPlayed",
                    description:
                      "Get tracks from the current user's recently played tracks. **Note**: Currently\ndoesn't support podcast episodes.",
                    args: [
                      {
                        name: "after",
                        description:
                          "A Unix timestamp in milliseconds. Returns all items after (but not\nincluding) this cursor position. If after is specified, before must not be\nspecified.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "before",
                        description:
                          "A Unix timestamp in milliseconds. Returns all items before (but not\nincluding) this cursor position. If before is specified, after must not be\nspecified.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "RecentlyPlayedConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Playlist",
                description:
                  "Information about a playlist owned by a Spotify user",
                fields: [
                  {
                    name: "id",
                    description:
                      "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the playlist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "collaborative",
                    description:
                      "`true` if the owner allows other users to modify the playlist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "description",
                    description:
                      "The playlist description. _Only returned for modified, verified playlists,\notherwise `null`_.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalUrls",
                    description: "Known external URLs for this playlist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExternalUrl",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "images",
                    description:
                      "Images for the playlist. The array may be empty or contain up to three images.\nThe images are returned by size in descending order.\nSee [Working with Playlists](https://developer.spotify.com/documentation/general/guides/working-with-playlists/).\n**Note**: If returned, the source URL for the image (`url`) is temporary and\nwill expire in less than a day.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Image",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "name",
                    description: "The name of the playlist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "owner",
                    description: "The user who owns the playlist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "User",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "public",
                    description:
                      "The playlist's public/private status: `true` the playlist is public, `false`\nthe playlist is private, `null` the playlist status is not relevant. For more\nabout public/private status, see [Working with Playlists](https://developer.spotify.com/documentation/general/guides/working-with-playlists/)",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "Boolean",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "tracks",
                    description: "The tracks of the playlist.",
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first item to return. Default: 0 (the first item).\n\nUse with `limit` to get the next set of items.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PlaylistTrackConnection",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) */\nfor the playlist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "PlaylistConnection",
                description: "A paged set of playlists",
                fields: [
                  {
                    name: "edges",
                    description: "The set of playlists.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "PlaylistEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for the set of playlists",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "PlaylistEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The playlist",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Playlist",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INTERFACE",
                name: "PlaylistTrack",
                description: null,
                fields: [
                  {
                    name: "durationMs",
                    description: "The playlist track length in milliseconds.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalUrls",
                    description: "External URLs for this episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExternalUrl",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint providing full details of the episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "id",
                    description:
                      "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for the playlist track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "name",
                    description: "The name of the episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the playlist track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: [
                  {
                    kind: "OBJECT",
                    name: "Episode",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                  {
                    kind: "OBJECT",
                    name: "Track",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                ],
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "PlaylistTrackConnection",
                description: "A paged set of tracks for a playlist",
                fields: [
                  {
                    name: "edges",
                    description:
                      "Pagination information for the tracks belonging to a playlist",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "PlaylistTrackEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for the tracks belonging to a playlist",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "PlaylistTrackEdge",
                description: null,
                fields: [
                  {
                    name: "addedAt",
                    description:
                      "The date and time the track was added to the playlist",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "DateTime",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "addedBy",
                    description:
                      "The user that added the track to the playlist",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "User",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "node",
                    description: "The playlist track",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "INTERFACE",
                        name: "PlaylistTrack",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Query",
                description: null,
                fields: [
                  {
                    name: "album",
                    description: "Spotify catalog information for an album.",
                    args: [
                      {
                        name: "id",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "Album",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "albums",
                    description:
                      "Get Spotify catalog information for multiple albums identified by their Spotify IDs.",
                    args: [
                      {
                        name: "ids",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "SCALAR",
                                name: "ID",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Album",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "artist",
                    description: "Spotify catalog information for an artist.",
                    args: [
                      {
                        name: "id",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "Artist",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "artists",
                    description:
                      "Get Spotify catalog information for several artists based on their Spotify IDs.",
                    args: [
                      {
                        name: "ids",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "SCALAR",
                                name: "ID",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Artist",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "developer",
                    description:
                      "Get a list of developer-specific settings, such as GraphQL field configuration.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Developer",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "episode",
                    description:
                      "Get Spotify catalog information for a single episode identified by its unique\nSpotify ID.",
                    args: [
                      {
                        name: "id",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "Episode",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "episodes",
                    description:
                      "Get Spotify catalog information for several episodes based on their Spotify IDs.",
                    args: [
                      {
                        name: "ids",
                        description:
                          "A list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for the episodes. Maximum: 50 IDs.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "SCALAR",
                                name: "ID",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Episode",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "genres",
                    description:
                      "A list of available genres seed parameter values for\n[recommendations](https://developer.spotify.com/documentation/web-api/reference/#/operations/get-recommendations).",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "String",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "recommendations",
                    description:
                      "Recommendations for the current user.\n\nRecommendations are generated based on the available information for a given\nseed entity and matched against similar artists and tracks. If there is\nsufficient information about the provided seeds, a list of tracks will be\nreturned together with pool size details.\n\nFor artists and tracks that are very new or obscure there might not be enough\ndata to generate a list of tracks.",
                    args: [
                      {
                        name: "seeds",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "RecommendationSeedInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "acousticness",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationAcousticnessInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "danceability",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationDanceabilityInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "durationMs",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationDurationMsInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "energy",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationEnergyInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "instrumentalness",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationInstrumentalnessInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "key",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationKeyInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "liveness",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationLivenessInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "loudness",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationLoudnessInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "mode",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationModeInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "popularity",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationPopularityInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "speechiness",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationSpeechinessInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "tempo",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationTempoInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "timeSignature",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationTimeSignatureInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "valence",
                        description: null,
                        type: {
                          kind: "INPUT_OBJECT",
                          name: "RecommendationValenceInput",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "limit",
                        description:
                          "The target size of the list of recommended tracks. For seeds with unusually\nsmall pools or when highly restrictive filtering is applied, it may be\nimpossible to generate the requested number of recommended tracks. Debugging\ninformation for such cases is available in the response.\n\nDefault value: 20.\nMinimum value: 1.\nMaximum value: 100.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "Recommendations",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "me",
                    description:
                      "Information about the current logged-in user.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "CurrentUser",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "newReleases",
                    description:
                      "Get a list of new album releases featured in Spotify (shown, for example, on a Spotify player’s “Browse” tab).",
                    args: [
                      {
                        name: "country",
                        description:
                          "A country: an [ISO 3166-1 alpha-2 country code](http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2). Provide this parameter if you want the list of returned items to be relevant to a particular country. If omitted, the returned items will be relevant to all countries.",
                        type: {
                          kind: "SCALAR",
                          name: "CountryCode",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first item to return. Default: 0 (the first item). Use with `limit` to get the next set of items.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "NewReleasesConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "playlist",
                    description: "A playlist owned by a Spotify user.",
                    args: [
                      {
                        name: "id",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "Playlist",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "featuredPlaylists",
                    description:
                      "A list of Spotify featured playlists (shown, for example, on a Spotify\nplayer's 'Browse' tab).",
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of items to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first playlist to return. Default: 0 (the first object).\n\nUse with `limit` to get the next set of playlists.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "timestamp",
                        description:
                          "Specify the local time to get results tailored for that specific date and\ntime in the day. If not provided, the response defaults to the current UTC\ntime. If there were no featured playlists (or there is no data) at the\nspecified time, the response will revert to the current UTC time.",
                        type: {
                          kind: "SCALAR",
                          name: "DateTime",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "FeaturedPlaylistConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "search",
                    description:
                      "Get Spotify catalog information about albums, artists, playlists, tracks, shows, episodes or audiobooks that match a keyword string.\n\n**Note: Audiobooks are only available for the US, UK, Ireland, New Zealand and Australia markets.**",
                    args: [
                      {
                        name: "q",
                        description:
                          "Your search query.\n\nYou can narrow down your search using field filters. The available filters are album, artist, track, year, upc, tag:hipster, tag:new, isrc, and genre. Each field filter only applies to certain result types.\n\nThe artist and year filters can be used while searching albums, artists and tracks. You can filter on a single year or a range (e.g. 1955-1960).\nThe album filter can be used while searching albums and tracks.\nThe genre filter can be used while searching artists and tracks.\nThe isrc and track filters can be used while searching tracks.\nThe upc, tag:new and tag:hipster filters can only be used while searching albums. The tag:new filter will return albums released in the past two weeks and tag:hipster can be used to return only albums with the lowest 10% popularity.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "String",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "includeExternal",
                        description:
                          "If `includeExternal=audio` is specified it signals that the client can play externally hosted audio content, and marks the content as playable in the response. By default externally hosted audio content is marked as unplayable in the response.",
                        type: {
                          kind: "ENUM",
                          name: "SearchExternalValue",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "limit",
                        description:
                          "The maximum number of results to return in each item type.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first result to return. Use with `limit` to get the next page of search results.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "type",
                        description:
                          "A list of item types to search across. Search results include hits from all the specified item types.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "ENUM",
                                name: "SearchType",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "SearchResults",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "show",
                    description:
                      "Get Spotify catalog information for a single show identified by its unique\nSpotify ID.",
                    args: [
                      {
                        name: "id",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "Show",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "shows",
                    description:
                      "Get Spotify catalog information for several shows based on their Spotify IDs.",
                    args: [
                      {
                        name: "ids",
                        description:
                          "A list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for the shows. Maximum: 50 IDs.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "SCALAR",
                                name: "ID",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Show",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "track",
                    description:
                      "Get Spotify catalog information for a single track identified by its unique\nSpotify ID.",
                    args: [
                      {
                        name: "id",
                        description:
                          "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the track.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "Track",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "tracks",
                    description:
                      "Get Spotify catalog information for multiple tracks based on their Spotify IDs.",
                    args: [
                      {
                        name: "ids",
                        description:
                          "A list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for the tracks. Maximum: 50 IDs.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "SCALAR",
                                name: "ID",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Track",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "tracksAudioFeatures",
                    description:
                      "Get audio features for multiple tracks based on their Spotify IDs.",
                    args: [
                      {
                        name: "ids",
                        description:
                          "A list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for the tracks. Maximum: 100 IDs.",
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "LIST",
                            name: null,
                            ofType: {
                              kind: "NON_NULL",
                              name: null,
                              ofType: {
                                kind: "SCALAR",
                                name: "ID",
                                ofType: null,
                                __typename: "IntrospectionType",
                              },
                              __typename: "IntrospectionType",
                            },
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "TrackAudioFeatures",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "user",
                    description: null,
                    args: [
                      {
                        name: "id",
                        description: null,
                        type: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "User",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "RecentlyPlayedConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of recently played items.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "RecentlyPlayedEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "RecentlyPlayedEdge",
                description: null,
                fields: [
                  {
                    name: "playedAt",
                    description: "The date and time the track was played at.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "DateTime",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "node",
                    description: "The item that was recently played.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "INTERFACE",
                        name: "PlaybackItem",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "context",
                    description: "The playback context for the track",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackContext",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationAcousticnessInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "SCALAR",
                name: "Float",
                description:
                  "The `Float` scalar type represents signed double-precision fractional values as specified by [IEEE 754](https://en.wikipedia.org/wiki/IEEE_floating_point).",
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationDanceabilityInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationDurationMsInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationEnergyInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationInstrumentalnessInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationKeyInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationLivenessInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationLoudnessInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationModeInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationPopularityInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Recommendations",
                description:
                  "Information about recommendations for the current user",
                fields: [
                  {
                    name: "seeds",
                    description:
                      "An array of recommendation [seed objects](https://developer.spotify.com/documentation/web-api/reference/#object-recommendationseedobject).",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "RecommendationSeed",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "tracks",
                    description:
                      "An array of [track object (simplified)](https://developer.spotify.com/documentation/web-api/reference/#object-simplifiedtrackobject)\nordered according to the parameters supplied.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "Track",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "RecommendationSeed",
                description:
                  "Information about a recommendation [seed object](https://developer.spotify.com/documentation/web-api/reference/#object-recommendationseedobject).",
                fields: [
                  {
                    name: "id",
                    description:
                      "The id used to select this seed. This will be the same as the string used in\nthe `seedArtists`, `seedTracks` or `seedGenres` parameter.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "afterFilteringSize",
                    description:
                      "The number of tracks available after min_* and max_* filters have been\napplied.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "afterRelinkingSize",
                    description:
                      "The number of tracks available after relinking for regional availability.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the full track or artist data for this seed. For tracks this will\nbe a link to a [Track Object](https://developer.spotify.com/documentation/web-api/reference/#object-trackobject).\nFor artists a link to an [Artist Object](https://developer.spotify.com/documentation/web-api/reference/#object-artistobject).\nFor genre seeds, this value will be `null`.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "initialPoolSize",
                    description:
                      "The number of recommended tracks available for this seed.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "type",
                    description: "The entity type of this seed.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "ENUM",
                        name: "RecommendationSeedType",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationSeedInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "seedArtists",
                    description:
                      'A list of [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for seed artists. Up to 5 seed values may be provided in any combination of `seedArtists`, `seedTracks` and `seedGenres`.\n\nExample value: ["4NHQUGzhtTLFvgF5SZesLK"]',
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "ID",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "seedGenres",
                    description:
                      'A list of any genres in the set of available genre seeds. Up to 5 seed values may be provided in any combination of `seedArtists`, `seedTracks` and `seedGenres`.\n\nExample value: ["classical", "country"]',
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "String",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "seedTracks",
                    description:
                      'A list of [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor a seed track. Up to 5 seed values may be provided in any combination of\n`seedArtists`, `seedTracks` and `seedGenres`.\n\nExample value: ["0c6xIDDpzE81m2q797ordA"]',
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "ID",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "RecommendationSeedType",
                description: "Available entity types for recommendation seeds.",
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "ARTIST",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "TRACK",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "GENRE",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationSpeechinessInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationTempoInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationTimeSignatureInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RecommendationValenceInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "min",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "max",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "target",
                    description: null,
                    type: {
                      kind: "SCALAR",
                      name: "Float",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ReleaseDate",
                description: null,
                fields: [
                  {
                    name: "date",
                    description:
                      "The date the item was first released, for example `1981-12-15`. Depending on\nthe precision, it might be shown as `1981-12`, or `1981-12-15`.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "precision",
                    description:
                      "The precision with which the `date` value is known.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "ENUM",
                        name: "ReleaseDatePrecision",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "ReleaseDatePrecision",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "YEAR",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "MONTH",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "DAY",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RemoveItemFromPlaylistInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "playlistId",
                    description:
                      "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nof the playlist.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "snapshotId",
                    description:
                      "The playlist's snapshot ID against which you want to make the changes. The API\nwill validate that the specified items exist and in the specified positions\nand make the changes, even if more recent changes have been made to the\nplaylist.",
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "tracks",
                    description:
                      "An array of objects containing [Spotify URIs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nof the tracks or episodes to remove.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "INPUT_OBJECT",
                            name: "RemoveItemFromPlaylistTrackInput",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "RemoveItemFromPlaylistPayload",
                description: null,
                fields: [
                  {
                    name: "snapshotId",
                    description: "A snapshot ID for the playlist",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "playlist",
                    description: "The playlist after the item was removed",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "Playlist",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RemoveItemFromPlaylistTrackInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "uri",
                    description: null,
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RemoveSavedAlbumsInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "ids",
                    description:
                      "A list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids).\nMaximum 20 IDs.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "RemoveSavedAlbumsPayload",
                description: null,
                fields: [
                  {
                    name: "removedAlbums",
                    description:
                      "The albums that were removed from the Spotify user's library.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Album",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RemoveSavedEpisodesInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "ids",
                    description:
                      "A list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids).\nMaximum 50 IDs.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "RemoveSavedEpisodesPayload",
                description: null,
                fields: [
                  {
                    name: "removedEpisodes",
                    description:
                      "The episodes that were removed from the Spotify user's library.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Episode",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RemoveSavedShowsInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "ids",
                    description:
                      "A list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids).\nfor the shows. Maximum 50 IDs.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "RemoveSavedShowsPayload",
                description: null,
                fields: [
                  {
                    name: "removedShows",
                    description:
                      "The shows that were removed from the Spotify user's library.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Show",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "RemoveSavedTracksInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "ids",
                    description:
                      "A comma-separated list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids).\nMaximum 50 IDs.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "RemoveSavedTracksPayload",
                description: null,
                fields: [
                  {
                    name: "removedTracks",
                    description:
                      "The tracks that were removed from the Spotify user's library.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Track",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "RepeatMode",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "CONTEXT",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "OFF",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "TRACK",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "ResetFieldConfigInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "field",
                    description:
                      "The field that will be reset to its default values",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "INPUT_OBJECT",
                        name: "FieldInput",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ResetFieldConfigPayload",
                description: null,
                fields: [
                  {
                    name: "fieldConfig",
                    description: "The updated field config",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "FieldConfig",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "ResumePlaybackInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "contextUri",
                    description:
                      "Spotify URI of the context to play. Valid contexts are albums, artists &\nplaylists.",
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "deviceId",
                    description:
                      "The id of the device this command is targeting. If not supplied, the user's\ncurrently active device is the target.",
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "offset",
                    description:
                      "Indicates from where in the context playback should start. Only available when\ncontextUri corresponds to an album or playlist object.",
                    type: {
                      kind: "INPUT_OBJECT",
                      name: "ResumePlaybackOffsetInput",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "positionMs",
                    description:
                      "Indicates the position where playback should occur in milliseconds.",
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "uris",
                    description: "An array of the Spotify track URIs to play.",
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "SCALAR",
                          name: "String",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "ResumePlaybackOffsetInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "position",
                    description:
                      "Non-negative, zero-based value that corresponds to the numeric position in the\nalbum or playlist",
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "uri",
                    description:
                      "Spotify URI of the item in the album or playlist",
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ResumePlaybackPayload",
                description: null,
                fields: [
                  {
                    name: "playbackState",
                    description: null,
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ResumePoint",
                description: null,
                fields: [
                  {
                    name: "fullyPlayed",
                    description:
                      "Whether or not the episode has been fully played by the user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "resumePositionMs",
                    description:
                      "The user's most recent position in the episode in milliseconds.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "SaveAlbumsInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "ids",
                    description:
                      "A list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the albums. Maximum: 20 IDs",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SaveAlbumsPayload",
                description: null,
                fields: [
                  {
                    name: "savedAlbums",
                    description:
                      "The albums that were saved to the Spotify user's library",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Album",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SavedAlbumEdge",
                description: null,
                fields: [
                  {
                    name: "addedAt",
                    description: "The date the album was saved.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "DateTime",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "node",
                    description: "The album object.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Album",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SavedAlbumsConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of saved albums.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "SavedAlbumEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for the set of playlists",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SavedEpisodeEdge",
                description: null,
                fields: [
                  {
                    name: "addedAt",
                    description: "The date the episode was saved.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "DateTime",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "node",
                    description: "The saved episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Episode",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SavedEpisodesConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of saved episodes.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "SavedEpisodeEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for the set of episodes",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SavedShowEdge",
                description: null,
                fields: [
                  {
                    name: "addedAt",
                    description: "The date the show was saved.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "DateTime",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "node",
                    description: "The show",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Show",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SavedShowsConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "A list of saved shows.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "SavedShowEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      '"Pagination information for the set of saved shows"',
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SavedTrackEdge",
                description: null,
                fields: [
                  {
                    name: "addedAt",
                    description: "The date the track was saved.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "DateTime",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "node",
                    description: "The track",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Track",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SavedTracksConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "A list of saved tracks.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "SavedTrackEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      '"Pagination information for the set of playlists"',
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "SaveEpisodesInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "ids",
                    description:
                      "An list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids).\nMaximum: 50 IDs",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SaveEpisodesPayload",
                description: null,
                fields: [
                  {
                    name: "savedEpisodes",
                    description:
                      "The episodes that were saved to the Spotify user's library",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Episode",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "SaveShowsInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "ids",
                    description:
                      "An list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids).\nfor the shows. Maximum: 50 IDs",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SaveShowsPayload",
                description: null,
                fields: [
                  {
                    name: "savedShows",
                    description:
                      "The shows that were saved to the Spotify user's library",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Show",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "SaveTracksInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "ids",
                    description:
                      "A list of the [Spotify IDs](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids).\nMaximum: 50 IDs",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SaveTracksPayload",
                description: null,
                fields: [
                  {
                    name: "savedTracks",
                    description:
                      "The tracks that were saved to the Spotify user's library",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Track",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SchemaField",
                description: null,
                fields: [
                  {
                    name: "typename",
                    description:
                      "The parent type name in the schema (ex: `User`)",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "fieldName",
                    description:
                      "The name of the field in the type (ex: `firstName`)",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "SchemaFieldInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "typename",
                    description:
                      "The parent type name in the schema (ex: `User`)",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "fieldName",
                    description:
                      "The name of the field in the type (ex: `firstName`)",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchAlbumEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The album returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Album",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchAlbumsConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of albums returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "SearchAlbumEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for albums in a search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchArtistEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The artist returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Artist",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchArtistsConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of artists returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "SearchArtistEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for artists in a search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchEpisodeEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The episode returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Episode",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchEpisodesConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description:
                      "The list of episodes returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "SearchEpisodeEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for episodes in a search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "SearchExternalValue",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "AUDIO",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchPlaylistEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The playlist returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Playlist",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchPlaylistsConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description:
                      "The list of playlists returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "SearchPlaylistEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for playlists in a search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchResults",
                description: null,
                fields: [
                  {
                    name: "albums",
                    description:
                      "The set of albums returned from the search query. Only available if the search `type` includes `ALBUM`.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "SearchAlbumsConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "artists",
                    description:
                      "The set of artists returned from the search query. Only available if the search `type` includes `ARTIST`.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "SearchArtistsConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "episodes",
                    description:
                      "The set of episodes returned from the search query. Only available if the search `type` includes `EPISODE`.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "SearchEpisodesConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "playlists",
                    description:
                      "The set of playlists returned from the search query. Only available if the search `type` includes `PLAYLIST`.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "SearchPlaylistsConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "shows",
                    description:
                      "The set of shows returned from the search query. Only available if the search `type` includes `SHOW`.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "SearchShowsConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "tracks",
                    description:
                      "The set of tracks returned from the search query. Only available if the search `type` includes `TRACK`.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "SearchTracksConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchShowEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The show returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Show",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchShowsConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of shows returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "SearchShowEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description: "Pagination information for shows in a search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchTrackEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The track returned in the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Track",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SearchTracksConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of tracks returned from the search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "SearchTrackEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for tracks in a search",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "SearchType",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "ALBUM",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "ARTIST",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "EPISODE",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "PLAYLIST",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "TRACK",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "SHOW",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "SeekToPositionContextInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "deviceId",
                    description:
                      "The id of the device this command is targeting. If not supplied, the user's currently active device is the target.",
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SeekToPositionResponse",
                description: null,
                fields: [
                  {
                    name: "playbackState",
                    description:
                      "The updated state of playback after seeking to a position.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "SetRepeatModeContextInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "deviceId",
                    description:
                      "The id of the device this command is targeting. If not supplied, the user's currently active device is the target.",
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SetRepeatModeResponse",
                description: null,
                fields: [
                  {
                    name: "playbackState",
                    description:
                      "The updated state of playback after setting a repeat mode.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "SetVolumeContextInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "deviceId",
                    description:
                      "The id of the device this command is targeting. If not supplied, the user's currently active device is the target.",
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SetVolumeResponse",
                description: null,
                fields: [
                  {
                    name: "playbackState",
                    description:
                      "The state of playback after the volume has been set.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Show",
                description: "Spotify catalog information for a show.",
                fields: [
                  {
                    name: "description",
                    description: "A description of the show.",
                    args: [
                      {
                        name: "format",
                        description: null,
                        type: {
                          kind: "ENUM",
                          name: "TextFormat",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: "PLAIN",
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "episodes",
                    description:
                      "Spotify catalog information about an show’s episodes.",
                    args: [
                      {
                        name: "limit",
                        description:
                          "The maximum number of episodes to return. Default: 20. Minimum: 1. Maximum: 50.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                      {
                        name: "offset",
                        description:
                          "The index of the first playlist to return. Default: 0 (the first object).\n\nUse with `limit` to get the next set of episodes.",
                        type: {
                          kind: "SCALAR",
                          name: "Int",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        defaultValue: null,
                        __typename: "IntrospectionInputValue",
                      },
                    ],
                    type: {
                      kind: "OBJECT",
                      name: "ShowEpisodesConnection",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "explicit",
                    description:
                      "Whether or not the show has explicit content (`true` = yes it does; `false`\n= no it does not OR unknown).",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalUrls",
                    description: "External URLs for this show.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExternalUrl",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint providing full details of the show.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "id",
                    description:
                      "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the show.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "images",
                    description:
                      "The cover art for the show in various sizes, widest first.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "Image",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "isExternallyHosted",
                    description:
                      "`true` if all of the shows episodes are hosted outside of Spotify's CDN. This\nfield might be `null` in some cases.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "Boolean",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "languages",
                    description:
                      "A list of the languages used in the show, identified by their [ISO 639](https://en.wikipedia.org/wiki/ISO_639) code.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "String",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "mediaType",
                    description: "The media type of the show.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "name",
                    description: "The name of the episode.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "publisher",
                    description: "The publisher of the show.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for the show.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ShowEpisodeEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The episode",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Episode",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ShowEpisodesConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "A list of episodes for the show.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "ShowEpisodeEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for the set of episodes",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "ShufflePlaybackContextInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "deviceId",
                    description:
                      "The id of the device this command is targeting. If not supplied, the user's\ncurrently active device is the target.",
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "ShufflePlaybackResponse",
                description: null,
                fields: [
                  {
                    name: "playbackState",
                    description:
                      "The state of playback after shuffling playback.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "SkipToNextContextInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "deviceId",
                    description:
                      "The id of the device this command is targeting. If not supplied, the user's\ncurrently active device is the target.",
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SkipToNextResponse",
                description: null,
                fields: [
                  {
                    name: "playbackState",
                    description:
                      "The updated state of playback after skipping to next.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "SkipToPreviousContextInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "deviceId",
                    description:
                      "The id of the device this command is targeting. If not supplied, the user's\ncurrently active device is the target.",
                    type: {
                      kind: "SCALAR",
                      name: "ID",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "SkipToPreviousResponse",
                description: null,
                fields: [
                  {
                    name: "playbackState",
                    description:
                      "The updated state of playback after skipping to previous.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Subscription",
                description: null,
                fields: [
                  {
                    name: "playbackStateChanged",
                    description: null,
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "TextFormat",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "PLAIN",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "HTML",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "ENUM",
                name: "TimeRange",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: [
                  {
                    name: "LONG_TERM",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "MEDIUM_TERM",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                  {
                    name: "SHORT_TERM",
                    description: null,
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionEnumValue",
                  },
                ],
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "SCALAR",
                name: "Timestamp",
                description: null,
                fields: null,
                inputFields: null,
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "TopArtistEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The artist.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Artist",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "TopArtistsConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of top tracks.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "TopArtistEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for the set of top tracks.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "TopTrackEdge",
                description: null,
                fields: [
                  {
                    name: "node",
                    description: "The track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Track",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "TopTracksConnection",
                description: null,
                fields: [
                  {
                    name: "edges",
                    description: "The list of top tracks.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "TopTrackEdge",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "pageInfo",
                    description:
                      "Pagination information for the set of top tracks.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "PageInfo",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "Track",
                description: "Spotify catalog information for a track.",
                fields: [
                  {
                    name: "id",
                    description:
                      "The [Spotify ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for the track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "album",
                    description: "The album on which the track appears.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Album",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "artists",
                    description: "The artists who performed the track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "OBJECT",
                            name: "Artist",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "audioFeatures",
                    description: "The track's audio feature information",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "TrackAudioFeatures",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "discNumber",
                    description:
                      "The disc number (usually `1` unless the album consists of more than one disc).",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "durationMs",
                    description: "The track length in milliseconds",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "explicit",
                    description:
                      "Whether or not the track has explicit lyrics (`true` = yes it does;\n`false` = no it does not OR unknown)",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalIds",
                    description: "Known external IDs for the track.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "TrackExternalIds",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalUrls",
                    description: "Known external URLs for this track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExternalUrl",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint providing full details of the track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "isLocal",
                    description:
                      "Whether or not the track is from a local file.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "isPlayable",
                    description:
                      "Part of the response when [Track Relinking](https://developer.spotify.com/documentation/general/guides/track-relinking-guide/)\nis applied. If `true`, the track is playable in the given market.\nOtherwise `false`.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "name",
                    description: "The name of the track",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "popularity",
                    description:
                      "The popularity of the track. The value will be between 0 and 100, with 100\nbeing the most popular.\n\nThe popularity of a track is a value between 0 and 100, with 100 being the\nmost popular. The popularity is calculated by algorithm and is based, in the\nmost part, on the total number of plays the track has had and how recent those\nplays are.\n\nGenerally speaking, songs that are being played a lot now will have a higher\npopularity than songs that were played a lot in the past. Duplicate tracks\n(e.g. the same track from a single and an album) are rated independently.\nArtist and album popularity is derived mathematically from track popularity.\nNote: the popularity value may lag actual popularity by a few days: the value\nis not updated in real time.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "previewUrl",
                    description:
                      "A link to a 30 second preview (MP3 format) of the track. Can be `null`",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "trackNumber",
                    description:
                      "The number of the track. If an album has several discs, the track number is\nthe number on the specified disc.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor the track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [
                  {
                    kind: "INTERFACE",
                    name: "PlaylistTrack",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                  {
                    kind: "INTERFACE",
                    name: "PlaybackItem",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                ],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "TrackAudioFeatures",
                description: null,
                fields: [
                  {
                    name: "acousticness",
                    description:
                      "A confidence measure from 0.0 to 1.0 of whether the track is acoustic. 1.0 represents high confidence the track is acoustic.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Float",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "analysisUrl",
                    description:
                      "A URL to access the full audio analysis of this track. An access token is required to access this data.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "danceability",
                    description:
                      "Danceability describes how suitable a track is for dancing based on a combination of musical elements including tempo, rhythm stability, beat strength, and overall regularity. A value of 0.0 is least danceable and 1.0 is most danceable.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Float",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "durationMs",
                    description: "The duration of the track in milliseconds.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "energy",
                    description:
                      "Energy is a measure from 0.0 to 1.0 and represents a perceptual measure of intensity and activity. Typically, energetic tracks feel fast, loud, and noisy. For example, death metal has high energy, while a Bach prelude scores low on the scale. Perceptual features contributing to this attribute include dynamic range, perceived loudness, timbre, onset rate, and general entropy.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Float",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "id",
                    description: "The Spotify ID for the track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "instrumentalness",
                    description:
                      'Predicts whether a track contains no vocals. "Ooh" and "aah" sounds are treated as instrumental in this context. Rap or spoken word tracks are clearly "vocal". The closer the instrumentalness value is to 1.0, the greater likelihood the track contains no vocal content. Values above 0.5 are intended to represent instrumental tracks, but confidence is higher as the value approaches 1.0.',
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Float",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "key",
                    description:
                      "The key the track is in. Integers map to pitches using standard [Pitch Class notation](https://en.wikipedia.org/wiki/Pitch_class). E.g. 0 = C, 1 = C♯/D♭, 2 = D, and so on. If no key was detected, the value is -1.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "liveness",
                    description:
                      "Detects the presence of an audience in the recording. Higher liveness values represent an increased probability that the track was performed live. A value above 0.8 provides strong likelihood that the track is live.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Float",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "loudness",
                    description:
                      "The overall loudness of a track in decibels (dB). Loudness values are averaged across the entire track and are useful for comparing relative loudness of tracks. Loudness is the quality of a sound that is the primary psychological correlate of physical strength (amplitude). Values typically range between -60 and 0 db.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Float",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "mode",
                    description:
                      "Mode indicates the modality (major or minor) of a track, the type of scale from which its melodic content is derived. Major is represented by 1 and minor is 0.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "speechiness",
                    description:
                      "Speechiness detects the presence of spoken words in a track. The more exclusively speech-like the recording (e.g. talk show, audio book, poetry), the closer to 1.0 the attribute value. Values above 0.66 describe tracks that are probably made entirely of spoken words. Values between 0.33 and 0.66 describe tracks that may contain both music and speech, either in sections or layered, including such cases as rap music. Values below 0.33 most likely represent music and other non-speech-like tracks.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Float",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "tempo",
                    description:
                      "The overall estimated tempo of a track in beats per minute (BPM). In musical terminology, tempo is the speed or pace of a given piece and derives directly from the average beat duration.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Float",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "timeSignature",
                    description:
                      'An estimated time signature. The time signature (meter) is a notational convention to specify how many beats are in each bar (or measure). The time signature ranges from 3 to 7 indicating time signatures of "3/4", to "7/4".',
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Int",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "trackHref",
                    description:
                      "A link to the Web API endpoint providing full details of the track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description: "The Spotify URI for the track.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "valence",
                    description:
                      "A measure from 0.0 to 1.0 describing the musical positiveness conveyed by a track. Tracks with high valence sound more positive (e.g. happy, cheerful, euphoric), while tracks with low valence sound more negative (e.g. sad, depressed, angry).",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Float",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "TrackExternalIds",
                description: null,
                fields: [
                  {
                    name: "isrc",
                    description:
                      "[International Standard Recording Code](http://en.wikipedia.org/wiki/International_Standard_Recording_Code)",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "ean",
                    description:
                      "[International Article Number](http://en.wikipedia.org/wiki/International_Article_Number_%28EAN%29)",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "upc",
                    description:
                      "[Universal Product Code](http://en.wikipedia.org/wiki/Universal_Product_Code)",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "TransferPlaybackInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "deviceIds",
                    description:
                      "A list containing the ID of the device on which playback should be\nstarted/transferred.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "LIST",
                        name: null,
                        ofType: {
                          kind: "NON_NULL",
                          name: null,
                          ofType: {
                            kind: "SCALAR",
                            name: "ID",
                            ofType: null,
                            __typename: "IntrospectionType",
                          },
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "play",
                    description:
                      "`true`: ensure playback happens on new device.\n`false` or not provided: keep the current playback state.",
                    type: {
                      kind: "SCALAR",
                      name: "Boolean",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "TransferPlaybackPayload",
                description: null,
                fields: [
                  {
                    name: "playbackState",
                    description:
                      "The state of playback after transferring devices.",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "PlaybackState",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INPUT_OBJECT",
                name: "UpdateFieldConfigInput",
                description: null,
                fields: null,
                inputFields: [
                  {
                    name: "config",
                    description: null,
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "INPUT_OBJECT",
                        name: "FieldConfigInput",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "field",
                    description: null,
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "INPUT_OBJECT",
                        name: "FieldInput",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                interfaces: null,
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "UpdateFieldConfigPayload",
                description: null,
                fields: [
                  {
                    name: "fieldConfig",
                    description: "The updated field config",
                    args: [],
                    type: {
                      kind: "OBJECT",
                      name: "FieldConfig",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "OBJECT",
                name: "User",
                description: "Public profile information about a Spotify user.",
                fields: [
                  {
                    name: "id",
                    description:
                      "The [Spotify user ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for this user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "displayName",
                    description:
                      "The name displayed on the user's profile. `null` if not available.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "externalUrls",
                    description: "Known public external URLs for this user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "ExternalUrl",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "followers",
                    description:
                      "Information about the followers of this user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Followers",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint for this user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "images",
                    description: "The user's profile image.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Image",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor this user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [
                  {
                    kind: "INTERFACE",
                    name: "UserProfile",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                ],
                enumValues: null,
                possibleTypes: null,
                __typename: "IntrospectionType",
              },
              {
                kind: "INTERFACE",
                name: "UserProfile",
                description: null,
                fields: [
                  {
                    name: "id",
                    description:
                      "The [Spotify user ID](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids) for this user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "ID",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "displayName",
                    description:
                      "The name displayed on the user's profile. `null` if not available.",
                    args: [],
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "followers",
                    description:
                      "Information about the followers of this user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "OBJECT",
                        name: "Followers",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "href",
                    description:
                      "A link to the Web API endpoint for this user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "images",
                    description: "The user's profile image.",
                    args: [],
                    type: {
                      kind: "LIST",
                      name: null,
                      ofType: {
                        kind: "NON_NULL",
                        name: null,
                        ofType: {
                          kind: "OBJECT",
                          name: "Image",
                          ofType: null,
                          __typename: "IntrospectionType",
                        },
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                  {
                    name: "uri",
                    description:
                      "The [Spotify URI](https://developer.spotify.com/documentation/web-api/#spotify-uris-and-ids)\nfor this user.",
                    args: [],
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    isDeprecated: false,
                    deprecationReason: null,
                    __typename: "IntrospectionField",
                  },
                ],
                inputFields: null,
                interfaces: [],
                enumValues: null,
                possibleTypes: [
                  {
                    kind: "OBJECT",
                    name: "CurrentUserProfile",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                  {
                    kind: "OBJECT",
                    name: "User",
                    ofType: null,
                    __typename: "IntrospectionType",
                  },
                ],
                __typename: "IntrospectionType",
              },
            ],
            directives: [
              {
                name: "synthetics",
                description: null,
                locations: ["FIELD"],
                args: [
                  {
                    name: "timeout",
                    description:
                      "The synthetic timeout configured for the field.",
                    type: {
                      kind: "SCALAR",
                      name: "Int",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "errorRate",
                    description:
                      "The synthetic error rate configured for the field.",
                    type: {
                      kind: "SCALAR",
                      name: "ErrorRate",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                  {
                    name: "enabled",
                    description: "Enable or disable synthetics for the field.",
                    type: {
                      kind: "SCALAR",
                      name: "Boolean",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: "true",
                    __typename: "IntrospectionInputValue",
                  },
                ],
                __typename: "IntrospectionDirective",
              },
              {
                name: "include",
                description:
                  "Directs the executor to include this field or fragment only when the `if` argument is true.",
                locations: ["FIELD", "FRAGMENT_SPREAD", "INLINE_FRAGMENT"],
                args: [
                  {
                    name: "if",
                    description: "Included when true.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                __typename: "IntrospectionDirective",
              },
              {
                name: "skip",
                description:
                  "Directs the executor to skip this field or fragment when the `if` argument is true.",
                locations: ["FIELD", "FRAGMENT_SPREAD", "INLINE_FRAGMENT"],
                args: [
                  {
                    name: "if",
                    description: "Skipped when true.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "Boolean",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                __typename: "IntrospectionDirective",
              },
              {
                name: "deprecated",
                description:
                  "Marks an element of a GraphQL schema as no longer supported.",
                locations: [
                  "FIELD_DEFINITION",
                  "ARGUMENT_DEFINITION",
                  "INPUT_FIELD_DEFINITION",
                  "ENUM_VALUE",
                ],
                args: [
                  {
                    name: "reason",
                    description:
                      "Explains why this element was deprecated, usually also including a suggestion for how to access supported similar data. Formatted using the Markdown syntax, as specified by [CommonMark](https://commonmark.org/).",
                    type: {
                      kind: "SCALAR",
                      name: "String",
                      ofType: null,
                      __typename: "IntrospectionType",
                    },
                    defaultValue: '"No longer supported"',
                    __typename: "IntrospectionInputValue",
                  },
                ],
                __typename: "IntrospectionDirective",
              },
              {
                name: "specifiedBy",
                description:
                  "Exposes a URL that specifies the behavior of this scalar.",
                locations: ["SCALAR"],
                args: [
                  {
                    name: "url",
                    description:
                      "The URL that specifies the behavior of this scalar.",
                    type: {
                      kind: "NON_NULL",
                      name: null,
                      ofType: {
                        kind: "SCALAR",
                        name: "String",
                        ofType: null,
                        __typename: "IntrospectionType",
                      },
                      __typename: "IntrospectionType",
                    },
                    defaultValue: null,
                    __typename: "IntrospectionInputValue",
                  },
                ],
                __typename: "IntrospectionDirective",
              },
            ],
            __typename: "IntrospectionSchema",
          },
          __typename: "Schema",
        },
      },
    },
  },
};
const GetSchemaByTag_WRONG_TOKEN = {
  operationName: "GetSchemaByTag",
  variables: { id: "spotify-demo-graph-519427f5", tag: "main" },
  response: {
    data: { service: null },
    errors: [
      {
        message: "HTTP fetch failed from 'kotlin': 406: Not Acceptable",
        extensions: {
          code: "SUBREQUEST_HTTP_ERROR",
          service: "kotlin",
          reason: "406: Not Acceptable",
          http: { status: 406 },
        },
      },
      { message: "Invalid credentials provided" },
    ],
  },
};

module.exports = {
  loadDefaultMocks,
  sendMock,
  FrontendUrlRoot,
  SchemaTagsAndFieldStats_NOT_ALLOWED_BY_USER_ROLE,
  SchemaTagsAndFieldStats,
  SchemaTagsAndFieldStats_WRONG_TOKEN,
  GetSchemaByTag,
  GetSchemaByTag_WRONG_TOKEN,
};
